package common

import (
	"bufio"
	"crypto/md5"
	"encoding/csv"
	"encoding/json"
	"fmt"
	"log"
	"net"
	"net/http"
	"strings"
	"sync"
	"time"
)

// Service defines a generic inventory service
type Service struct {
	ID     string `yaml:"id"`
	IP     string `yaml:"ip"`
	Port   int    `yaml:"port"`
	Volume string `yaml:"volume"`
	Type   string
	Conf   map[string]string `yaml:"config"`
}

type Metric struct {
	Name   string
	Labels [][]string
	Value  float64
	// Optional accumulator functionality
	Accumulator *Accumulator
	Type        string
	TTL         time.Duration
}

type MetricData struct {
	sync.Mutex
	Measures map[string]*Metric
}

type TCPCommandOutput struct {
	Key   string
	Value interface{}
}

type accValue struct {
	value float64
	ts    time.Time
}

type Accumulator struct {
	sync.Mutex
	ttl        time.Duration
	values     []*accValue
	outputType string
}

func (acc *Accumulator) Values() (res map[string]float64) {
	res = map[string]float64{}
	// Cleanup old values
	values := []float64{}
	acc.Lock()
	for _, v := range acc.values {
		if time.Since(v.ts) > acc.ttl {
			continue
		}
		values = append(values, v.value)
	}
	acc.Unlock()
	if len(values) == 0 {
		values = append(values, float64(0))
	}
	operations := []string{"sum"}
	if acc.outputType == "stats" {
		operations = []string{"max", "avg", "min"}
	}

	max, min, sum := acc.Compute(values)
	for _, op := range operations {
		switch op {
		case "sum":
			res["sum"] = sum
		case "max":
			res["max"] = max
		case "avg":
			res["avg"] = sum / float64(len(values))
		case "min":
			res["min"] = min
		}
	}
	return
}

func (acc *Accumulator) Compute(values []float64) (max, min, sum float64) {
	min = float64(1e20)
	for _, v := range values {
		sum += v
		if v < min {
			min = v
		}
		if v > max {
			max = v
		}
	}
	return
}

func (acc *Accumulator) Push(value float64) {
	acc.values = append(acc.values, &accValue{
		ts:    time.Now(),
		value: value,
	})
}

func MakeMetricData() *MetricData {
	return &MetricData{
		Measures: make(map[string]*Metric),
	}
}

func (m *MetricData) IncrError(module, serviceType string) {
	m.Acc(&Metric{
		Name: "self_errors",
		Labels: [][]string{
			[]string{"module", module},
			[]string{"service_type", serviceType},
		},
		Value: 1,
		Type:  "counter",
		TTL:   time.Duration(10) * time.Second,
	})
}

func (m *MetricData) fingerprint(metric Metric) string {
	payload := metric.Name
	for _, l := range metric.Labels {
		payload += l[1]
	}
	return fmt.Sprintf("%x", md5.Sum([]byte(payload)))
}

func (m *MetricData) Set(metric *Metric) {
	m.Lock()
	fp := m.fingerprint(*metric)
	m.Measures[fp] = metric
	m.Unlock()
}

func (m *MetricData) Acc(metric *Metric) {
	if metric.TTL == time.Duration(0) || metric.Type == "" {
		log.Println("WARN", "in accumulator", "missing TTL/Type init values", metric)
		return
	}
	m.Lock()
	fp := m.fingerprint(*metric)
	if _, ok := m.Measures[fp]; !ok {
		metric.Accumulator = &Accumulator{
			ttl:        metric.TTL,
			outputType: metric.Type,
			values:     []*accValue{},
		}
		m.Measures[fp] = metric
	}
	m.Measures[fp].Accumulator.Push(metric.Value)
	m.Unlock()
}

func HTTPGetJson(url string, into interface{}) error {
	resp, err := http.Get(url)
	if err != nil {
		return err
	}
	return json.NewDecoder(resp.Body).Decode(into)
}

func HTTPGetCsv(url string, sep rune) ([][]string, error) {
	resp, err := http.Get(url)
	if err != nil {
		return nil, err
	}
	r := csv.NewReader(resp.Body)
	r.Comma = sep
	return r.ReadAll()
}

func TCPCommand(target, cmd string, sep string, offset int, timeout time.Duration) (data []TCPCommandOutput, err error) {
	conn, err := net.Dial("tcp", target)
	if err != nil {
		return
	}
	defer conn.Close()
	if conn.SetDeadline(time.Now().Add(timeout)) != nil {
		log.Println("WARN", "in setting deadline for cmd", target, cmd)
		return
	}
	if _, err = conn.Write([]byte(cmd)); err != nil {
		return
	}
	scanner := bufio.NewScanner(conn)
	for scanner.Scan() {
		line := scanner.Text()
		kv := strings.Split(line, sep)
		if kv[0] == "END" || line == "" || line == "NOT_FOUND" {
			break
		}
		if len(kv) != 2+offset {
			continue
		}
		data = append(data, TCPCommandOutput{Key: kv[offset], Value: kv[1+offset]})
	}
	conn.Close()
	if err = scanner.Err(); err != nil {
		return
	}
	return
}

func Index(data []string, key string) int {
	for i, v := range data {
		if key == v {
			return i
		}
	}
	return -1
}
