package healthcheck

import (
	"context"
	"encoding/base64"
	"fmt"
	"github.com/open-io/oio-exporter/common"
	"gopkg.in/yaml.v2"
	"io/ioutil"
	"log"
	"net"
	"net/http"
	"os"
	"regexp"
	"strings"
	"time"
)

type daemon struct {
	targets      **[]*common.Service
	healthchecks map[string]healthcheck
	interval     time.Duration
}

type module struct {
	Name        string            `yaml:"name"`
	Prober      string            `yaml:"prober"`
	TimeoutStr  string            `yaml:"timeout"`
	Send        string            `yaml:"send"`
	ExpectStr   string            `yaml:"expect"`
	Method      string            `yaml:"method"`
	Headers     map[string]string `yaml:"headers"`
	StatusCodes []int             `yaml:"status_codes"`
	Route       string            `yaml:"route"`
	Timeout     time.Duration     `yaml:"_disabled"`
	Expect      *regexp.Regexp    `yaml:"_disabled2"`
}

type healthcheck interface {
	check(common.Service) (bool, error)
}

// HTTPCheck Defines a generic HTTP check
type HTTPCheck struct {
	*module
}

func (hc *HTTPCheck) check(target common.Service) (passed bool, err error) {
	req, err := http.NewRequest(hc.Method, "http://"+target.IP+fmt.Sprintf(":%d", target.Port)+hc.Route, strings.NewReader(hc.Send))
	if err != nil {
		return false, err
	}
	ctx, cancel := context.WithTimeout(context.Background(), hc.Timeout)
	defer cancel()

	resp, err := http.DefaultClient.Do(req.WithContext(ctx))
	if err != nil {
		return false, err
	}
	// Check for valid status codes
	passed = false
	for _, code := range hc.StatusCodes {
		if code == resp.StatusCode {
			passed = true
			break
		}
	}
	if !passed {
		return passed, nil
	}
	if hc.Expect == nil {
		return true, nil
	}
	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return false, err
	}
	return hc.Expect.Match(body), nil
}

// TCPCheck Defines a generic TCP check
type TCPCheck struct {
	*module
}

func (hc *TCPCheck) check(target common.Service) (passed bool, err error) {
	conn, err := net.Dial("tcp", fmt.Sprintf(target.IP+":%d", target.Port))
	if err != nil {
		return false, err
	}
	defer conn.Close()

	if err := conn.SetDeadline(time.Now().Add(hc.Timeout)); err != nil {
		return false, err
	}

	if hc.Send == "" && hc.Expect == nil {
		return true, nil
	}

	if _, err = conn.Write([]byte(hc.Send)); err != nil {
		return false, err
	}
	var res = make([]byte, 1e6)

	if _, err := conn.Read(res); err != nil {
		return false, err
	}
	return hc.Expect.Match(res), nil
}

func (d *daemon) parseHealthchecks(path string, checkStr string) (n int, err error) {
	var hc struct {
		Modules []*module
	}

	if path == "" {
		var data []byte

		data, err = base64.StdEncoding.DecodeString(checkStr)
		if err != nil {
			return
		}

		err = yaml.Unmarshal(data, &hc)
		if err != nil {
			return
		}
		log.Println("Healthchecks definition loaded from defaults")
	} else {
		var f *os.File
		f, err = os.Open(path)
		if err != nil {
			return
		}
		err = yaml.NewDecoder(f).Decode(&hc)
		f.Close()
		if err != nil {
			return
		}
		log.Printf("Healthchecks definition loaded from %s", path)
	}

	for _, module := range hc.Modules {
		var err error

		if module.Prober == "none" {
			d.healthchecks[module.Name] = nil
			n++
			continue
		}

		ok := true
		if module.Timeout, err = time.ParseDuration(module.TimeoutStr); err != nil {
			log.Printf(`WARN failed to parse 'timeout' for module %s: "%s"`, module.Name, err)
			ok = false
		}
		if module.ExpectStr != "" {
			if module.Expect, err = regexp.Compile(module.ExpectStr); err != nil {
				log.Printf(`WARN failed to parse 'expect' for module %s: "%s"`, module.Name, err)
				ok = false
			}
		}
		if module.Prober == "http" {
			d.healthchecks[module.Name] = &HTTPCheck{module}
		} else if module.Prober == "tcp" {
			d.healthchecks[module.Name] = &TCPCheck{module}
		} else {
			log.Printf(`WARN unknown 'prober' for module %s: "%s"`, module.Name, module.Prober)
			ok = false
		}

		if !ok {
			log.Printf("WARN ignoring module %s", module.Name)
			continue
		}

		n++
	}
	return
}

func MakeDaemon(hc string, healthcheckPath string, targets **[]*common.Service, interval time.Duration) *daemon {
	d := daemon{
		interval:     interval,
		targets:      targets,
		healthchecks: map[string]healthcheck{},
	}
	n, err := d.parseHealthchecks(healthcheckPath, hc)
	if err != nil {
		log.Fatalln("FATAL", "healthchecks loading failed", err)
	}
	log.Println("Loaded", n, "healthchecks")
	return &d
}

func (d *daemon) Run(m *common.MetricData, once bool) {
	for {
		if d.targets == nil {
			log.Println("No targets loaded")
			time.Sleep(d.interval)
			continue
		}
		for _, target := range **d.targets {
			if hc, ok := d.healthchecks[target.Type]; ok {
				if hc == nil {
					continue
				}
				val := float64(0)
				passed, _ := hc.check(*target)
				if passed {
					val = float64(1)
				}
				m.Set(&common.Metric{
					Name: "probe_success",
					Labels: [][]string{
						[]string{"service", target.ID},
						[]string{"service_type", target.Type},
						[]string{"addr", fmt.Sprintf(target.IP+":%d", target.Port)},
					},
					Value: val,
				})
			} else {
				log.Println("WARN", "unknown healthcheck", target.Type)
			}
		}
		if once {
			return
		}
		time.Sleep(d.interval)
	}
}
