package healthcheck

import (
	"encoding/base64"
	"fmt"
	"github.com/open-io/oio-exporter/common"
	"io/ioutil"
	"log"
	"net"
	"net/http"
	"strconv"
	"strings"
	"testing"
	"time"
)

func loadHC() string {
	res, err := ioutil.ReadFile("../healthchecks.yml")
	if err != nil {
		log.Fatalln(err)
	}
	return base64.StdEncoding.EncodeToString(res)
}

func makeFakeHTTPService(route, data string, statusCode int) (port int64) {
	listener, err := net.Listen("tcp", "127.0.0.1:0")
	if err != nil {
		log.Fatalln(err)
	}
	http.HandleFunc(route, func(w http.ResponseWriter, r *http.Request) {
		w.WriteHeader(statusCode)
		fmt.Fprint(w, data)
	})
	go func(listener net.Listener) {
		log.Fatalln(http.Serve(listener, nil))
	}(listener)

	addr := listener.Addr().String()

	port, err = strconv.ParseInt(strings.Split(addr, ":")[1], 10, 64)
	if err != nil {
		log.Fatalln(err)
	}

	return port
}

func runFakeTCPService(l net.Listener, data string) {
	for {
		conn, err := l.Accept()
		if err != nil {
			log.Printf("WARN: %s", err)
			return
		}
		go func(conn net.Conn) {
			defer conn.Close()
			var res []byte
			_, err := conn.Read(res)
			if err != nil {
				log.Fatalln(err)
			}
			_, err = conn.Write([]byte(data))
			if err != nil {
				log.Fatalln(err)
			}
		}(conn)
	}
}

func makeFakeTCPService(data string) (port int64) {
	listener, err := net.Listen("tcp", "127.0.0.1:0")
	if err != nil {
		log.Fatalln(err)
	}
	go runFakeTCPService(listener, data)

	addr := listener.Addr().String()

	port, err = strconv.ParseInt(strings.Split(addr, ":")[1], 10, 64)
	if err != nil {
		log.Fatalln(err)
	}

	return port
}

func makeFakeServices(d *daemon) map[string]float64 {
	expected := map[string]float64{}
	targets := []*common.Service{}
	targets = append(targets, &common.Service{
		ID:     "OPENIO-account-0",
		IP:     "127.0.0.1",
		Port:   int(makeFakeHTTPService(d.healthchecks["account"].(*HTTPCheck).Route, "", 200)),
		Volume: "",
		Type:   "account",
		Conf:   map[string]string{},
	})
	expected["OPENIO-account-0"] = float64(1)

	targets = append(targets, &common.Service{
		ID:     "OPENIO-meta0-0",
		IP:     "127.0.0.1",
		Port:   int(makeFakeTCPService("OK \n")),
		Volume: "",
		Type:   "meta0",
		Conf:   map[string]string{},
	})
	expected["OPENIO-meta0-0"] = float64(1)

	targets = append(targets, &common.Service{
		ID:     "OPENIO-beanstalkd-0",
		IP:     "127.0.0.1",
		Port:   int(makeFakeTCPService("OK \n")),
		Volume: "",
		Type:   "beanstalkd",
		Conf:   map[string]string{},
	})
	expected["OPENIO-beanstalkd-0"] = float64(1)

	targets = append(targets, &common.Service{
		ID:     "OPENIO-beanstalkd-1",
		IP:     "127.0.0.1",
		Port:   int(makeFakeTCPService("KO \n")),
		Volume: "",
		Type:   "beanstalkd",
		Conf:   map[string]string{},
	})
	expected["OPENIO-beanstalkd-1"] = float64(0)

	targets = append(targets, &common.Service{
		ID:     "OPENIO-keystone-0",
		IP:     "127.0.0.1",
		Port:   int(makeFakeHTTPService(d.healthchecks["keystone"].(*HTTPCheck).Route, "", 401)),
		Volume: "",
		Type:   "keystone",
		Conf:   map[string]string{},
	})
	expected["OPENIO-keystone-0"] = float64(1)

	targetPtr := &targets
	d.targets = &targetPtr
	return expected
}

func getLabelValue(data [][]string, key string) (value string) {
	for _, label := range data {
		if label[0] == key {
			return label[1]
		}
	}
	return ""
}

func TestHealtcheckModule(t *testing.T) {
	m := common.MakeMetricData()
	d := MakeDaemon(loadHC(), "", nil, time.Duration(30)*time.Second)
	expected := makeFakeServices(d)
	d.Run(m, true)
	for _, metric := range m.Measures {
		svc := getLabelValue(metric.Labels, "service")
		if svc == "" {
			t.Fatal("No label 'service' found", metric)
		}
		if val, ok := expected[svc]; ok {
			if metric.Value != val {
				t.Fatal("Healthcheck output error, expected", val, "got", metric.Value, svc)
			}
		} else {
			t.Fatal("Cannot determine expected result for service", svc)
		}
	}
}
