package openio

import (
	"fmt"
	"github.com/open-io/oio-exporter/common"
	"log"
	"os"
	"strconv"
	"strings"
	"syscall"
	"time"
)

type daemon struct {
	services **map[string][]*common.Service
	interval time.Duration
	proxyURL *string
	ns       string
	types    []string
}

type metaxInfoBody struct {
	Cache     map[string]int64
	Elections map[string]int64
}

type serviceInfoBody struct {
	Addr  string
	Score int
}

func MakeDaemon(ns string, services **map[string][]*common.Service, interval time.Duration) *daemon {
	// TODO: consider using lastUpdate to reload proxy url here
	var proxyURL string
	if svc, ok := (**services)["oioproxy"]; ok && len(svc) > 0 {
		proxyURL = fmt.Sprintf("http://%s:%d/v3.0/", svc[0].IP, svc[0].Port)
	}
	var types []string
	if err := common.HTTPGetJson(proxyURL+ns+"/conscience/info?what=types", &types); err != nil {
		log.Println("WARN", "failed to fetch service types from conscience", err)
	}

	d := daemon{
		ns:       ns,
		services: services,
		interval: interval,
		proxyURL: &proxyURL,
		types:    types,
	}
	return &d
}

func (d *daemon) Run(m *common.MetricData) {
	for {
		for serviceType, services := range **d.services {
			if common.Index(d.types, serviceType) == -1 {
				continue
			}
			d.collectScore(serviceType, services, m)
			for _, svc := range services {
				d.collectVolumeInfo(*svc, m)
				switch serviceType {
				case "meta0", "meta1", "meta2":
					d.collectMeta(*svc, m)
				case "rawx":
					d.collectRawx(*svc, m)
				}
			}
		}
		time.Sleep(d.interval)
	}
}

func (d *daemon) collectMeta(svc common.Service, m *common.MetricData) {
	if d.proxyURL == nil {
		return
	}
	// TODO: move target as service address
	target := fmt.Sprintf(svc.IP+":%d", svc.Port)

	counters, err := common.HTTPGetCsv(*d.proxyURL+"forward/stats?id="+target, ' ')
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN: metaX stats collection failed", err)
		return
	}
	for _, cntr := range counters {
		val, err := strconv.ParseFloat(cntr[2], 64)
		if err != nil {
			continue
		}
		if cntr[0] == "counter" {
			parts := strings.Split(cntr[1], ".")
			if len(parts) != 3 {
				continue
			}
			m.Set(&common.Metric{
				Name: parts[0] + "_" + parts[1],
				Labels: [][]string{
					[]string{"dimension", parts[2]},
					[]string{"service", target},
					[]string{"service_type", svc.Type},
					[]string{"addr", target},
					// TODO: add volume
				},
				Value: val,
			})
		}
	}
	svcInfo := metaxInfoBody{}
	err = common.HTTPGetJson(*d.proxyURL+"forward/info?id="+target, &svcInfo)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting meta2 info", svc.IP, svc.Port, err)
		return
	}
	for t, metrics := range map[string]map[string]int64{"cache": svcInfo.Cache, "elections": svcInfo.Elections} {
		for dim, val := range metrics {
			m.Set(&common.Metric{
				Name: t,
				Labels: [][]string{
					[]string{"dimension", dim},
					[]string{"service", target},
					[]string{"service_type", svc.Type},
					[]string{"addr", target},
				},
				Value: float64(val),
			})
		}
	}
}

func (d *daemon) collectRawx(svc common.Service, m *common.MetricData) {
	target := fmt.Sprintf(svc.IP+":%d", svc.Port)
	counters, err := common.HTTPGetCsv("http://"+target+"/stat", ' ')
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN: rawx metric collection failed", err)
		return
	}
	for _, cntr := range counters {
		val, err := strconv.ParseFloat(cntr[2], 64)
		if err != nil {
			continue
		}
		if cntr[0] == "counter" {
			parts := strings.Split(cntr[1], ".")
			if len(parts) != 3 {
				continue
			}
			m.Set(&common.Metric{
				Name: parts[0] + "_" + parts[1],
				Labels: [][]string{
					[]string{"dimension", parts[2]},
					[]string{"service", target},
					[]string{"service_type", svc.Type},
					[]string{"addr", target},
					// TODO: add volume
				},
				Value: val,
			})
		}
	}
}

func (d *daemon) collectScore(svcType string, services []*common.Service, m *common.MetricData) {
	info := []serviceInfoBody{}
	err := common.HTTPGetJson(*d.proxyURL+d.ns+"/conscience/list?type="+svcType, &info)
	if err != nil {
		m.IncrError("openio", svcType)
		log.Println("WARN", "when collecting score info", svcType, err)
		return
	}
	for _, svc := range services {
		target := fmt.Sprintf(svc.IP+":%d", svc.Port)
		for _, remoteSvc := range info {
			if remoteSvc.Addr == target {
				m.Set(&common.Metric{
					Name: "score",
					Labels: [][]string{
						[]string{"service", target},
						[]string{"service_type", svc.Type},
						[]string{"addr", target},
						// TODO: add volume
					},
					Value: float64(remoteSvc.Score),
				})
				break
			}
		}
	}
}

func (d *daemon) collectVolumeInfo(svc common.Service, m *common.MetricData) {
	if svc.Volume == "" {
		return
	}
	target := fmt.Sprintf(svc.IP+":%d", svc.Port)
	f, err := os.OpenFile(svc.Volume, os.O_RDONLY, 0644)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting volume info", svc, err)
		return
	}
	defer f.Close()

	var stat syscall.Statfs_t
	err = syscall.Fstatfs(int(f.Fd()), &stat)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting volume info", svc, err)
		return
	}
	vMetric := map[string]uint64{
		"byte_avail":  stat.Bavail * uint64(stat.Bsize),
		"byte_used":   (stat.Blocks - stat.Bfree) * uint64(stat.Bsize),
		"byte_free":   stat.Bfree * uint64(stat.Bsize),
		"inodes_free": stat.Ffree,
		"inodes_used": stat.Files - stat.Ffree,
	}
	fsID := strconv.FormatUint(uint64(stat.Fsid.X__val[1])<<32|uint64(stat.Fsid.X__val[0]), 10)
	for dim, val := range vMetric {
		m.Set(&common.Metric{
			Name: dim,
			Labels: [][]string{
				[]string{"service", target},
				[]string{"service_type", svc.Type},
				[]string{"addr", target},
				[]string{"volume", svc.Volume},
				[]string{"fsid", fsID},
			},
			Value: float64(val),
		})
	}
}
