package version

import (
	"encoding/base64"
	"github.com/open-io/oio-exporter/common"
	"gopkg.in/yaml.v2"
	"log"
	"os"
	"os/exec"
	"regexp"
	"strings"
	"time"
)

type daemon struct {
	interval   time.Duration
	packageCmd string
	packager   string
	versions   []*version
}

type version struct {
	Name    string         `yaml:"name"`
	Type    string         `yaml:"type"`
	Package string         `yaml:"package"`
	Rpm     string         `yaml:"rpm"`
	Deb     string         `yaml:"deb"`
	Command string         `yaml:"command"`
	Extract string         `yaml:"extract"`
	Regex   *regexp.Regexp `yaml:"_disabled2"`
}

func (d *daemon) parseVersions(path, defaultVersions string) (err error) {
	var versions struct {
		Versions []*version
	}

	if path == "" {
		var data []byte

		data, err = base64.StdEncoding.DecodeString(defaultVersions)
		if err != nil {
			return
		}

		err = yaml.Unmarshal(data, &versions)
		if err != nil {
			return
		}
		log.Println("Versions definition loaded from defaults")
	} else {
		var f *os.File
		f, err = os.Open(path)
		if err != nil {
			return
		}
		err = yaml.NewDecoder(f).Decode(&versions)
		f.Close()
		if err != nil {
			return
		}
		log.Printf("Versions definition loaded from %s", path)
	}

	for _, version := range versions.Versions {

		if version.Name == "" {
			log.Printf("WARN name is missing in one of the version definition")
			continue
		} else if version.Type == "package" {
			if version.Package == "" {
				if version.Deb == "" || version.Rpm == "" {
					log.Printf("ERR package type must declare package name as package, rpm and deb for %s", version.Name)
					continue
				}
			}
		} else if version.Type == "command" {
			if version.Command == "" {
				log.Printf("ERR command must be set for %s", version.Name)
				continue
			}
			if version.Extract != "" {
				version.Regex, err = regexp.Compile(version.Extract)
				if err != nil {
					log.Printf("ERR error extracting regex (%s) for %s: %s", version.Extract, version.Name, err)
					continue
				}
			}
		} else {
			log.Printf("WARN Wrong type (%s) for %s", version.Type, version.Name)
			continue
		}
		d.versions = append(d.versions, version)
	}
	//d.versions = versions.Versions
	return nil
}

func MakeDaemon(versions string, versionPath string, interval time.Duration) *daemon {
	d := daemon{
		interval:   interval,
		packageCmd: "",
		packager:   "",
		versions:   []*version{},
	}

	if checkCMD("rpm") {
		d.packageCmd = "rpm -q --qf %{VERSION}"
		d.packager = "rpm"
	} else if checkCMD("dpkg-query") {
		d.packageCmd = "dpkg-query -Wf=${Version}"
		d.packager = "deb"
	} else {
		log.Println("WARN", "in determining distro", "packages version will not be available")
	}

	err := d.parseVersions(versionPath, versions)
	if err != nil {
		log.Println("WARN", "Could not parse versions", err)
	}
	log.Printf("INFO: loaded %d versions", len(d.versions))

	return &d
}

func (d *daemon) Run(m *common.MetricData) {
	for {
		for _, version := range d.versions {

			var cmd string
			var regex *regexp.Regexp = nil

			if version.Type == "package" {
				if d.packageCmd == "" {
					continue
				}

				cmd = d.packageCmd + " "
				if version.Package != "" {
					cmd += version.Package
				} else if d.packager == "rpm" && version.Rpm != "" {
					cmd += version.Rpm
				} else if d.packager == "deb" && version.Deb != "" {
					cmd += version.Deb
				} else {
					continue
				}
			} else if version.Type == "command" {
				cmd = version.Command
				regex = version.Regex
			} else {
				continue
			}
			out, err := d.runCommand(cmd, regex)
			if err != nil {
				m.IncrError("command", version.Name)
				log.Println("WARN", "in running command", cmd, version.Name, err)
				continue
			}
			m.Set(&common.Metric{
				Name: "version",
				Labels: [][]string{
					[]string{"service_type", version.Name},
					[]string{"version", out},
				},
				Value: float64(time.Now().Unix()),
			})
		}
		time.Sleep(d.interval)
	}
}

func checkCMD(cmd string) bool {
	_, err := exec.LookPath(cmd)
	return err == nil
}

func (d *daemon) runCommand(cmd string, regex *regexp.Regexp) (string, error) {
	cmds := strings.Split(cmd, " ")
	out, err := exec.Command(cmds[0], cmds[1:]...).CombinedOutput()

	if err != nil {
		return "", err
	}

	outFmt := strings.Trim(strings.TrimSuffix(string(out), "\n"), " ")

	if regex == nil {
		return outFmt, nil
	}

	matches := regex.FindStringSubmatch(outFmt)
	if len(matches) < 1 {
		return "", nil
	}
	if len(matches) == 1 {
		return matches[0], nil
	}

	return matches[1], nil
}
